//OpenMP implemented RandomAccess test computing GUPS

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>


#include <omp.h>

/* Define 64-bit types and corresponding format strings for printf() and constants */
#ifdef LONG_IS_64BITS
typedef unsigned long u64Int;
typedef long s64Int;
#else
typedef unsigned long long u64Int;
typedef long long s64Int;
#endif

#define STREQ(a,b) (strcmp((a),(b)) == 0)

u64Int RA_starts(s64Int n);
void Access(u64Int* Table, u64Int TableLength,u64Int ithread, u64Int memorySizePerThread);

bool atomicity = false;
const u64Int POLY = 0x0000000000000007, PERIOD = 1317624576693539401L;

int main(int argc, char *args[] )
{
	// threads count 
	int threads = 1; 
	// memory size per thread
	u64Int memorySizePerThread;
	// use locking for atomic updates 
	atomicity = false;
	// verifying after test
	bool verify = true;

	if ((argc == 1) || (argc > 5))
	{
		printf("Usage: RandomAccessBenchmark.exe <threads> <mem_per_thread> [/atom] [/notest]\n");
		return 0;
	} 
	
	threads=(int)atoi(args[1]);
	memorySizePerThread = powl(2, atoi(args[2]));

	if (argc == 5)
	{
		atomicity = STREQ(args[3],"/atom") ? true : false;
		verify = STREQ(args[4],"/notest")  ? false : true;
	}
	if (argc == 4)
	{
		atomicity = STREQ(args[3],"/atom")  ? true : false;
		verify = STREQ(args[3],"/notest") ? false : true;
	}
	if (argc == 3)
	{
		if (STREQ(args[2],"/atom"))
			atomicity = true;
		else
			if (STREQ(args[2],"/notest"))
				verify = false;
	}

	printf("RandomAccessBenchmark\n");

	// totam memory size in use
	u64Int TableLength = threads * memorySizePerThread; 

	// total memory updates count
	u64Int updates_total = 4 * TableLength;

	printf("Threads in use = %d\n" , threads);
	printf("Total memory in use (in bytes) = %ld\n" , TableLength * sizeof(u64Int));
	printf("Table length = %ld\n" , TableLength);
	printf("Total updates count to be done = %ld\n" , updates_total);
	printf("Atomicity is %s\n" , (atomicity ? "on" : "off"));
	printf("Verifying is %s\n" , (verify ? "on" : "off"));
	printf("Please wait...\n");

	
	u64Int* Table = new u64Int[TableLength];
	for (u64Int k = 0; k < TableLength; k++)
		Table[k] = k;

	double start_time = omp_get_wtime();
	        
	//      thread team
	omp_set_dynamic(0);
//for(u64Int i=0;i<threads;i++)	
	#pragma omp parallel num_threads(threads)
	{
		
		int i=omp_get_thread_num();
		Access(Table, TableLength, i, memorySizePerThread);

		
	}
	
	double finish_time = omp_get_wtime();

	//double time_res_in_seconds = ((now.wMinute-old.wMinute) * 60 + (now.wSecond-old.wSecond) + (now.wMilliseconds-old.wMilliseconds) * 0.001);
	double time_res_in_seconds = finish_time - start_time;
	
	printf("CPU time used  = %f seconds\n" , time_res_in_seconds);
	printf("GUPS = %f\n", (updates_total / time_res_in_seconds / 1e9));

	if (verify)
	{
		printf("Verifying...\n");
		//for verification doing access in serial and safe mode
		Access(Table, TableLength, 0, TableLength);

		long errcount = 0;
		for (u64Int i = 0; i < TableLength; i++)
		{
			if(Table[i]!=i)
				errcount++;
		}
		printf("Found %d errors in %lu locations. (%f%%).\n" , errcount, TableLength, (float)errcount / (float)TableLength * 100.0f);
	}
	delete[] Table;
	return 0;

}
/* Utility routine to start random number generator at Nth step */
u64Int RA_starts(s64Int n)
{
  int i, j;
  u64Int m2[64];
  u64Int temp, ran;

  while (n < 0) n += PERIOD;
  while (n > PERIOD) n -= PERIOD;
  if (n == 0) return 0x1;

  temp = 0x1;
  for (i=0; i<64; i++) {
    m2[i] = temp;
    temp = (temp << 1) ^ ((s64Int) temp < 0 ? POLY : 0);
    temp = (temp << 1) ^ ((s64Int) temp < 0 ? POLY : 0);
  }

  for (i=62; i>=0; i--)
    if ((n >> i) & 1)
      break;

  ran = 0x2;
  while (i > 0) {
    temp = 0;
    for (j=0; j<64; j++)
      if ((ran >> j) & 1)
        temp ^= m2[j];
    ran = temp;
    i -= 1;
    if ((n >> i) & 1)
      ran = (ran << 1) ^ ((s64Int) ran < 0 ? POLY : 0);
  }

  return ran;
}

// Random Access and update memory
void Access(u64Int* Table, u64Int TableLength,u64Int ithread, u64Int memorySizePerThread)
{
	// memory updates number per thread
	u64Int updates = 4 * memorySizePerThread;

	// conpute "base" for pseudorandom numbers generation
	u64Int ran = RA_starts((s64Int)(updates * ithread ));

	for (u64Int i = 0; i < updates; i++)
	{
		if (atomicity)
		#pragma omp atomic
			Table[ran & (TableLength - 1)] ^= ran;
		else
			Table[ran & (TableLength - 1)] ^= ran;

		ran = (ran << 1) ^ ((s64Int)ran < 0 ? POLY : 0);
	}
}

